      module xml_node

      use strings

      integer, parameter :: ELEMENT_NODE            =  1
      integer, parameter :: ATTRIBUTE_NODE          =  2
      integer, parameter :: TEXT_NODE               =  3
      integer, parameter :: CDATA_SECTION_NODE      =  4
      integer, parameter :: ENTITY_REFERENCE_NODE   =  5
      integer, parameter :: ENTITY_NODE             =  6
      integer, parameter :: PI_NODE                 =  7
      integer, parameter :: COMMENT_NODE            =  8
      integer, parameter :: DOCUMENT_NODE           =  9
      integer, parameter :: DTD_NODE                = 10
      integer, parameter :: DOCUMENT_FRAGMENT_NODE  = 11
      integer, parameter :: NOTATION_NODE           = 12

      type node
!#ifndef IFC
!          private
!#endif
          integer              :: type = 0
          type(string)         :: name
          type(string)         :: value
          type(node), pointer  :: parent => null()

!***      nodenamemap  :: attributes
          integer              :: n_attributes = 0
          type(node), pointer  :: first_attribute => null()
          type(node), pointer  :: last_attribute => null()

!***      nodelist  :: children
          integer              :: n_children = 0
          type(node), pointer  :: first_child => null()
          type(node), pointer  :: last_child => null()

          type(node), pointer  :: previous_sibling => null()
          type(node), pointer  :: next_sibling => null()

          type(node), pointer  :: root => null()
      end type node


      contains

!*******************************************************************************

      function node_type(n)

      implicit none
      type(node), intent(in)  :: n
      integer                 :: node_type


      node_type = n%type

      end function node_type

!*******************************************************************************

      function get_root(n)

      implicit none
      type(node), intent(in)  :: n
      type(node), pointer     :: get_root


      get_root => n%root

      end function get_root

!*******************************************************************************

      function has_children(n)

      implicit none
      type(node), intent(in)  :: n
      logical                 :: has_children


      has_children = associated(n%first_child)

      end function has_children

!*******************************************************************************

      subroutine append_child(parent,type,name,value)

      implicit none
      type(node), pointer              :: parent
      integer, intent(in)              :: type
      character(*), intent(in)         :: name,value
      type(node), pointer, save        :: child



!*** This is a hack; need a special routine actually to
!*** initialize new documents
      if (type == DOCUMENT_NODE) then
          allocate(parent)
          parent%type = DOCUMENT_NODE
          return
      endif

      allocate(child)

      child%type = type
      child%name = name
      child%value = value

      child%parent => parent

      if (associated(parent%last_child)) then
          parent%last_child%next_sibling => child
      endif

     if (associated(parent%last_child)) then
         child%previous_sibling => parent%last_child
     endif

      if (.not. associated(parent%first_child)) then
          parent%first_child => child
      endif
      parent%last_child => child

      parent%n_children = parent%n_children + 1

      end subroutine append_child

!*******************************************************************************

      subroutine add_attribute(parent,name,value)

      implicit none
      type(node), pointer              :: parent
      character(*), intent(in)         :: name,value
      type(node), pointer, save        :: attribute


      allocate(attribute)

      attribute%type = ATTRIBUTE_NODE
      attribute%name = name
      attribute%value = value

      attribute%parent => parent

      if (associated(parent%last_attribute)) then
          parent%last_attribute%next_sibling => attribute
      endif

     if (associated(parent%last_attribute)) then
         attribute%previous_sibling => parent%last_attribute
     endif

      if (.not. associated(parent%first_attribute)) then
          parent%first_attribute => attribute
      endif
      parent%last_attribute => attribute

      parent%n_attributes = parent%n_attributes + 1

      end subroutine add_attribute

!*******************************************************************************

      function get_parent(child)

      implicit none
      type(node), intent(in)           :: child
      type(node), pointer              :: get_parent


      if (associated(child%parent)) then
         get_parent => child%parent
      else
          get_parent => null()
      endif

      end function get_parent

!*******************************************************************************

      function first_child_node(parent)

      implicit none
      type(node), pointer              :: parent
      type(node), pointer              :: first_child_node


      if (associated(parent%first_child)) then
          first_child_node => parent%first_child
      else
          first_child_node => null()
      endif

      end function first_child_node

!*******************************************************************************

      function last_child_node(parent)

      implicit none
      type(node), pointer              :: parent
      type(node), pointer              :: last_child_node

      if (associated(parent%last_child)) then
          last_child_node => parent%last_child
      else
          last_child_node => null()
      endif

      end function last_child_node

!*******************************************************************************

      function next_sibling_node(n)

      implicit none
      type(node), pointer              :: n
      type(node), pointer              :: next_sibling_node


      if (associated(n%next_sibling)) then
          next_sibling_node => n%next_sibling
      else
          next_sibling_node => null()
      endif

      end function next_sibling_node

!*******************************************************************************

      function previous_sibling_node(n)

      implicit none
      type(node), pointer              :: n
      type(node), pointer              :: previous_sibling_node


      if (associated(n%previous_sibling)) then
          previous_sibling_node => n%previous_sibling
      else
          previous_sibling_node => null()
      endif

      end function previous_sibling_node

!*******************************************************************************

      recursive subroutine print_node(io,n)

      implicit none
      integer, intent(in)       :: io
      type(node), intent(in)    :: n
      character(*), parameter   :: f = '(a)'
      character(1)              :: d
      type(node), pointer       :: c


      select case(n%type)
      case(DOCUMENT_NODE)
          c => n%first_child
          do while(associated(c))
              call print_node(io,c)
              c => c%next_sibling
          enddo
      case(ELEMENT_NODE)
          write(io,f,advance='no') '<' // n%name
          if (associated(n%first_attribute)) then
              c => n%first_attribute
              do while(associated(c))
                  call print_node(io,c)
                  c => c%next_sibling
              enddo
          endif
          if (associated(n%first_child)) then
              write(io,f,advance='no') '>'
              c => n%first_child
              do while(associated(c))
                  call print_node(io,c)
                  c => c%next_sibling
              enddo
              write(io,f,advance='no') '</' // n%name // '>'
          else
              write(io,f,advance='no') '/>'
          endif
      case(DTD_NODE)
          write(io,f) '<!DOCTYPE ' // n%value // '>'
      case(ATTRIBUTE_NODE)
          if (index(n%value,'"') /= 0) then
              d = "'"
          else
              d = '"'
          endif
          write(io,f,advance='no') ' ' // n%name // '=' // d // n%value // d
      case(PI_NODE)
          if (n%parent%type == DOCUMENT_NODE) then
              write(io,f,advance='yes') '<?' // n%value // '>'
          else
              write(io,f,advance='no' ) '<?' // n%value // '>'
          endif
      case(COMMENT_NODE)
          if (n%parent%type == DOCUMENT_NODE) then
              write(io,f,advance='yes') '<!--' // n%value // '-->'
          else
              write(io,f,advance='no' ) '<!--' // n%value // '-->'
          endif
      case(CDATA_SECTION_NODE)
          write(io,f,advance='no') '<![CDATA[' // n%value // ']]>'
      case(TEXT_NODE)
          write(io,f,advance='no') char(n%value)
      case(ENTITY_REFERENCE_NODE)
          write(io,f,advance='no') '&' // n%value // ';'
      case(ENTITY_NODE)
          write(io,f,advance='no') '&' // n%value // ';'
      case(DOCUMENT_FRAGMENT_NODE)
          continue
      case(NOTATION_NODE)
          continue
      case default
          write(6,*) 'PRINT_NODE: unknown node type', n%type
          stop
      end select

      end subroutine print_node

!*******************************************************************************

      end module xml_node
